#!/usr/bin/env node

'use strict';

var commander = require('commander');
var path = require('path');
var rxjs = require('rxjs');
var operators = require('rxjs/operators');
var fs = require('fs');
var inquirer = require('inquirer');

function _interopNamespace(e) {
    if (e && e.__esModule) return e;
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () {
                        return e[k];
                    }
                });
            }
        });
    }
    n['default'] = e;
    return Object.freeze(n);
}

var path__namespace = /*#__PURE__*/_interopNamespace(path);
var fs__namespace = /*#__PURE__*/_interopNamespace(fs);
var inquirer__namespace = /*#__PURE__*/_interopNamespace(inquirer);

const isBrowser = () => typeof window !== 'undefined';

// NOTE: use bind(window) to avoid `TypeError: Illegal invocation`
const fetch = isBrowser() ? globalThis.fetch.bind(window) : require('node-fetch');
class DefaultHttpClient {
    fetchFn;
    constructor(fetchFn = fetch) {
        this.fetchFn = fetchFn;
    }
    async get(url, options = {}) {
        const result = await this.fetchFn(url, options);
        return (await result.json());
    }
}

const baseURL = 'https://scrapbox.io/api';
const isObject = (data) => typeof data === 'object' && data !== null;
const isNotLoggedInError = (data) => isObject(data) && data['name'] === 'NotLoggedInError';
class RestApi {
    token;
    httpClient;
    constructor(token, httpClient = new DefaultHttpClient()) {
        this.token = token;
        this.httpClient = httpClient;
    }
    async getPage(projectName, pageName) {
        return this.request(`${baseURL}/pages/${encodeURIComponent(projectName)}/${encodeURIComponent(pageName)}?followRename=true`);
    }
    async getProject(projectName) {
        return this.request(`${baseURL}/projects/${encodeURIComponent(projectName)}`);
    }
    async getProjects() {
        return this.request(`${baseURL}/projects`);
    }
    async getMe() {
        return this.request(`${baseURL}/users/me`);
    }
    async request(url) {
        const result = await this.httpClient.get(url, {
            headers: {
                Cookie: `connect.sid=${this.token}`,
            },
        });
        if (isNotLoggedInError(result)) {
            throw new Error('NotLoggedInError');
        }
        return result;
    }
}

const packetTypes = {
    initialize: '0',
    ping: '2',
    pong: '3',
    connected: '40',
    // e.g.
    //   * 42: 'cursor'
    //   * 42X: user custom request (X is arbitrary natural number to specify response)
    send: '42',
    // e.g.
    //   * 43X: response to user custom request (X is number used for request)
    receive: '43',
    response: '43',
};
const CONFIG = {
    endpoint: 'wss://scrapbox.io/socket.io/?EIO=4&transport=websocket',
    origin: 'https://scrapbox.io',
    responseTimeout: 1000 * 15,
};

const isIntegerString = (value) => Number.isInteger(Number.parseInt(value));
const getAuthCookieValue = (token) => `connect.sid=${token}`;

const toSocketIoMessagePayload = (id, data) => {
    return `${packetTypes.send}${id}${JSON.stringify(['socket.io-request', data])}`;
};
const isConnectionMessage = (message) => {
    return message[0] === packetTypes.initialize;
};
const isResponseOf = (expected) => ({ sid }) => expected === sid;
// 430[{...}}] => 430, [{}]
// @see https://github.com/socketio/engine.io-protocol
const scrapboxDeserializer = (event) => {
    let packetType = '';
    let message = String(event.data);
    while (message.length) {
        const head = message[0];
        if (isIntegerString(head)) {
            packetType += head;
            message = message.substr(1);
        }
        else {
            return [packetType, JSON.parse(message)];
        }
    }
    return [packetType, undefined];
};
const scrapboxSerializer = (message) => {
    return typeof message === 'string' ? message : JSON.stringify(message);
};

// copied from official index.js
// modified original toString16 of official index.js
const toString16 = (userIdOrNumber, length) => {
    const hexOrId = userIdOrNumber.toString(16);
    return hexOrId.length > length ? hexOrId.slice(hexOrId.length - length) : '0'.repeat(length - hexOrId.length) + hexOrId;
};
const map = new Map();
const getIdGenerator = (userId) => {
    if (map.has(userId)) {
        return map.get(userId);
    }
    let counter = Math.floor(16777215 * Math.random());
    const generate = () => {
        const t = Math.floor(Date.now() / 1e3);
        if ((counter += 1) > 16777215) {
            counter = 0;
        }
        return (toString16(t, 8) + toString16(userId, 6) + toString16(0, 4) + toString16(counter, 6));
    };
    map.set(userId, generate);
    return generate;
};
const generateId = (userId) => {
    const generator = getIdGenerator(userId);
    return generator();
};

// NOTE: userId is used to generate new ID
const createChanges = (params, userId) => params.map((param) => {
    switch (param.type) {
        case 'insert': {
            return { _insert: param.position || '_end', lines: { id: generateId(userId), text: param.text } };
        }
        case 'update': {
            return { _update: param.id, lines: { text: param.text } };
        }
        case 'delete': {
            return { _delete: param.id, lines: -1 };
        }
        case 'title': {
            return { title: param.title };
        }
        case 'description': {
            return { descriptions: [param.text] };
        }
        default: {
            throw new Error('Bad implement');
        }
    }
});

const defaultSerializer = (data) => JSON.stringify(data);
const defaultDeserializer = (message) => message.data;
const isMessageEvent = (event) => event.type === 'message';
class RxWebSocket {
    websocket;
    opened$;
    events$ = new rxjs.Subject();
    serializer = defaultSerializer;
    deserializer = defaultDeserializer;
    constructor({ url, protocols, options, serializer = defaultSerializer, deserializer = defaultDeserializer, }) {
        const websocket = isBrowser() ? new WebSocket(url, protocols) : new (require('ws'))(url, protocols, options);
        const close$ = rxjs.fromEvent(websocket, 'close').pipe(rxjs.take(1), operators.share());
        const open$ = rxjs.fromEvent(websocket, 'open').pipe(operators.takeUntil(close$), rxjs.shareReplay(1), rxjs.take(1));
        const message$ = rxjs.fromEvent(websocket, 'message').pipe(operators.takeUntil(close$), rxjs.shareReplay(1));
        const error$ = rxjs.fromEvent(websocket, 'error').pipe(operators.takeUntil(close$), rxjs.mergeMap(rxjs.throwError));
        rxjs.merge(open$, message$, error$, close$).subscribe(this.events$);
        this.websocket = websocket;
        this.opened$ = open$;
        this.serializer = serializer.bind(this);
        this.deserializer = deserializer.bind(this);
    }
    get active() {
        return !this.events$.closed;
    }
    get message() {
        return this.events$.asObservable().pipe(rxjs.filter(isMessageEvent), operators.map(this.deserializer));
    }
    send(message) {
        return this.opened$.pipe(rxjs.mergeMap(() => {
            this.websocket.send(this.serializer(message));
            return this.message;
        }));
    }
}

class ScrapboxWebsocketHandler {
    options;
    socket = null;
    room = null;
    sid = 0;
    constructor(options) {
        this.options = options;
    }
    async commit({ projectId, userId, pageId, parentId, changes, }) {
        await this.join({ projectId, pageId });
        return this.send({
            projectId,
            pageId,
            data: {
                method: 'commit',
                data: { kind: 'page', userId, projectId, pageId, parentId, changes: createChanges(changes, userId), cursor: null, freeze: true },
            },
        });
    }
    async join({ projectId, pageId }) {
        // join room if needed
        if (this.room?.projectId === projectId && this.room?.pageId === pageId) {
            return;
        }
        await this.send({
            projectId,
            pageId,
            data: {
                method: 'room:join',
                data: { pageId, projectId, projectUpdatesStream: false },
            },
        });
        this.room = { projectId, pageId };
    }
    async send({ data, projectId, pageId }) {
        const sid = String(this.sid++);
        this.getActiveSocket()
            .send(toSocketIoMessagePayload(sid, data))
            .pipe(operators.first(isResponseOf(sid)), operators.timeout(CONFIG.responseTimeout));
    }
    getActiveSocket() {
        return !this.socket?.active ? (this.socket = this.openNewSocket()) : this.socket;
    }
    openNewSocket() {
        const socket = new RxWebSocket({
            url: CONFIG.endpoint,
            options: { headers: { Origin: CONFIG.origin, Cookie: getAuthCookieValue(this.options.token ?? '') } },
            serializer: scrapboxSerializer,
            deserializer: scrapboxDeserializer,
        });
        socket.message.pipe(operators.first(isConnectionMessage)).subscribe(([, data]) => {
            socket.send(packetTypes.connected);
            rxjs.interval(data.pingInterval)
                .pipe(operators.takeUntil(socket.message))
                .subscribe(() => socket.send(packetTypes.ping));
        });
        return socket;
    }
}

class ScrapboxApi {
    restApi;
    websocketHandler;
    constructor(token = '', restApi = new RestApi(token), websocketHandler = new ScrapboxWebsocketHandler({ token })) {
        this.restApi = restApi;
        this.websocketHandler = websocketHandler;
    }
    async getPage(projectName, pageName) {
        return this.restApi.getPage(projectName, pageName);
    }
    async changeLine(projectName, pageName, change) {
        const project = await this.getProject(projectName);
        const { pageId, commitId } = await this.getPageIdAndCommitId(projectName, pageName);
        return this.commit({
            changes: Array.isArray(change) ? change : [change],
            projectId: project.id,
            pageId: pageId,
            commitId: commitId,
        });
    }
    async getUser() {
        return await this.restApi.getMe();
    }
    async getProject(projectName) {
        return await this.restApi.getProject(projectName);
    }
    async getPageIdAndCommitId(projectName, pageName) {
        const page = await this.restApi.getPage(projectName, pageName);
        return { pageId: page.id, commitId: page.commitId };
    }
    async commit(param) {
        const user = await this.getUser();
        return this.websocketHandler.commit({
            userId: user.id,
            projectId: param.projectId,
            pageId: param.pageId,
            parentId: param.commitId,
            changes: param.changes,
        });
    }
}

const loadFile = (path) => new Promise((resolve, reject) => fs__namespace.readFile(path, 'utf8', (err, data) => {
    if (err) {
        reject(err);
    }
    resolve(data);
}));
const getJSTString = () => new Date().toLocaleString('ja-JP', { timeZone: 'Asia/Tokyo' });
const versionComment = `/* deployed at: ${getJSTString()} */`;
// first space is needed for scrapbox code block
const loadSourceCode = async (file) => ` ${versionComment}${await loadFile(file)}`;

const findIndex = (searchString, lines) => {
    return lines.findIndex((line) => line.text.includes(searchString));
};
const findNextLineIdOrFail = (searchString, lines) => {
    const index = findIndex(searchString, lines);
    if (index === -1) {
        throw new Error('Line not found');
    }
    const nextLineOrNone = lines[index + 1];
    if (!nextLineOrNone) {
        throw new Error('Next line not found');
    }
    return nextLineOrNone.id;
};

const isNonEmptyString = (value) => typeof value === 'string' && value.length > 0;
const isCSSFile = (value) => value.endsWith('.css');
const isJSFile = (value) => value.endsWith('.js');
const validateDeployArguments = ({ project, page, sourceFilePath, token }) => {
    const errors = [];
    if (!isNonEmptyString(project)) {
        errors.push('project is required');
    }
    if (!isNonEmptyString(page)) {
        errors.push('page is required');
    }
    if (!isNonEmptyString(sourceFilePath)) {
        errors.push('filepath is required');
    }
    if (!(isCSSFile(sourceFilePath) || isJSFile(sourceFilePath))) {
        errors.push('file type is not css or js');
    }
    if (!isNonEmptyString(token)) {
        errors.push('token is required');
    }
    return errors;
};

const deploy = async ({ page, sourceFilePath, project, token }) => {
    const errors = validateDeployArguments({ token, sourceFilePath, project, page });
    if (errors.length) {
        throw new Error(`ParameterError: ${JSON.stringify(errors)}`);
    }
    const codeBlockLabel = isCSSFile(sourceFilePath) ? 'style.css' : 'script.js';
    const sourceCode = await loadSourceCode(sourceFilePath);
    const scrapboxClient = new ScrapboxApi(token);
    const pageData = await scrapboxClient.getPage(project, page);
    // NOTE: assuming code-block exists already
    const lineId = findNextLineIdOrFail(codeBlockLabel, pageData.lines);
    // FIXME: bad type assertion
    await scrapboxClient.changeLine(project, page, { type: 'update', id: lineId, text: sourceCode });
};

const selectProject = async ({ client }) => {
    const projectOptions = await client.getProjects().then(({ projects }) => projects.map(({ name }) => name));
    const answers = await inquirer__namespace.prompt([{ type: 'list', name: 'project', message: 'select target project', choices: projectOptions }]);
    return answers.project;
};

const pkg = require(path__namespace.resolve(__dirname, '../../package.json'));
const program = new commander.Command().version(pkg.version);
program
    .command('deploy-userscript <js-file>')
    .description('update userscript in user page')
    .option('-p, --project <type>', 'project name')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApi(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    const userName = await api.getMe().then(({ name }) => name);
    await deploy({ token, project: selectedProject, page: userName, sourceFilePath: jsFile });
});
program
    .command('deploy-usercss <css-file>')
    .description('update usercss in settings page')
    .option('-p, --project <type>', 'project name')
    .action(async (jsFile, options) => {
    const token = process.env.TOKEN ?? '';
    if (!token) {
        throw new Error('TOKEN not set');
    }
    const api = new RestApi(token);
    const selectedProject = options.project ?? (await selectProject({ client: api }));
    await deploy({ token, project: selectedProject, page: 'settings', sourceFilePath: jsFile });
});
program
    .parseAsync(process.argv)
    .then(() => {
    process.exit(0);
})
    .catch((e) => {
    console.error(e);
    process.exit(1);
});
//# sourceMappingURL=bin.js.map
