import { interval } from 'rxjs';
import { first, timeout, takeUntil } from 'rxjs/operators';
import { CONFIG, packetTypes } from './constants.js';
import { toSocketIoMessagePayload, isResponseOf, scrapboxSerializer, scrapboxDeserializer, isConnectionMessage } from './internal/message.js';
import { createChanges } from './internal/request.js';
import { getAuthCookieValue } from './internal/util.js';
import { RxWebSocket } from './internal/websocket.js';

class ScrapboxWebsocketHandler {
    options;
    socket = null;
    room = null;
    sid = 0;
    constructor(options) {
        this.options = options;
    }
    async commit({ projectId, userId, pageId, parentId, changes, }) {
        await this.join({ projectId, pageId });
        return this.send({
            projectId,
            pageId,
            data: {
                method: 'commit',
                data: { kind: 'page', userId, projectId, pageId, parentId, changes: createChanges(changes, userId), cursor: null, freeze: true },
            },
        });
    }
    async join({ projectId, pageId }) {
        // join room if needed
        if (this.room?.projectId === projectId && this.room?.pageId === pageId) {
            return;
        }
        await this.send({
            projectId,
            pageId,
            data: {
                method: 'room:join',
                data: { pageId, projectId, projectUpdatesStream: false },
            },
        });
        this.room = { projectId, pageId };
    }
    async send({ data, projectId, pageId }) {
        const sid = String(this.sid++);
        this.getActiveSocket()
            .send(toSocketIoMessagePayload(sid, data))
            .pipe(first(isResponseOf(sid)), timeout(CONFIG.responseTimeout));
    }
    getActiveSocket() {
        return !this.socket?.active ? (this.socket = this.openNewSocket()) : this.socket;
    }
    openNewSocket() {
        const socket = new RxWebSocket({
            url: CONFIG.endpoint,
            options: { headers: { Origin: CONFIG.origin, Cookie: getAuthCookieValue(this.options.token ?? '') } },
            serializer: scrapboxSerializer,
            deserializer: scrapboxDeserializer,
        });
        socket.message.pipe(first(isConnectionMessage)).subscribe(([, data]) => {
            socket.send(packetTypes.connected);
            interval(data.pingInterval)
                .pipe(takeUntil(socket.message))
                .subscribe(() => socket.send(packetTypes.ping));
        });
        return socket;
    }
}

export { ScrapboxWebsocketHandler };
//# sourceMappingURL=websocket-handler.js.map
